"""
Module to provide for a simple implementation of a title case algorithm.
"""
import argparse
import glob
import logging
import os
import runpy
import sys
import traceback

from pymarkdown.application_properties import (
    ApplicationProperties,
    ApplicationPropertiesJsonLoader,
)
from pymarkdown.bad_tokenization_error import BadTokenizationError
from pymarkdown.parser_logger import ParserLogger
from pymarkdown.plugin_manager import BadPluginError, PluginManager
from pymarkdown.source_providers import FileSourceProvider
from pymarkdown.tokenized_markdown import TokenizedMarkdown

POGGER = ParserLogger(logging.getLogger(__name__))

LOGGER = logging.getLogger(__name__)


class PyMarkdownLint:
    """
    Class to provide for a simple implementation of a title case algorithm.
    """

    available_log_maps = {
        "CRITICAL": logging.CRITICAL,
        "ERROR": logging.ERROR,
        "WARNING": logging.WARNING,
        "INFO": logging.INFO,
        "DEBUG": logging.DEBUG,
    }

    def __init__(self):
        self.__version_number = PyMarkdownLint.__get_semantic_version()
        self.__show_stack_trace = False

        self.__properties = ApplicationProperties()

        self.__plugins = PluginManager()
        self.__tokenizer = None
        self.default_log_level = "CRITICAL"

    @staticmethod
    def __get_semantic_version():
        file_path = __file__
        if not os.path.isabs(file_path):
            assert False
        file_path = file_path.replace(os.sep, "/")
        last_index = file_path.rindex("/")
        file_path = file_path[0 : last_index + 1] + "version.py"
        version_meta = runpy.run_path(file_path)
        return version_meta["__version__"]

    @staticmethod
    def log_level_type(argument):
        """
        Function to help argparse limit the valid log levels.
        """
        if argument in PyMarkdownLint.available_log_maps:
            return argument
        raise ValueError(f"Value '{argument}' is not a valid log level.")

    def __parse_arguments(self):
        parser = argparse.ArgumentParser(description="Lint any found Markdown files.")

        parser.add_argument(
            "-e",
            "--enable-rules",
            dest="enable_rules",
            action="store",
            default="",
            help="comma separated list of rules to enable",
        )
        parser.add_argument(
            "-d",
            "--disable-rules",
            dest="disable_rules",
            action="store",
            default="",
            help="comma separated list of rules to disable",
        )
        parser.add_argument(
            "-x-scan",
            dest="x_test_scan_fault",
            action="store_true",
            default="",
            help=argparse.SUPPRESS,
        )
        parser.add_argument(
            "-x-init",
            dest="x_test_init_fault",
            action="store_true",
            default="",
            help=argparse.SUPPRESS,
        )
        parser.add_argument(
            "--add-plugin",
            dest="add_plugin",
            action="append",
            default=None,
            help="path to a plugin containing a new rule to apply",
        )
        parser.add_argument(
            "--config",
            "-c",
            dest="configuration_file",
            action="store",
            default=None,
            help="path to the configuration file to use",
        )
        parser.add_argument(
            "--set",
            "-s",
            dest="set_configuration",
            action="append",
            default=None,
            help="manually set an individual configuration property",
            type=ApplicationProperties.verify_manual_property_form,
        )
        parser.add_argument(
            "--strict-config",
            dest="strict_configuration",
            action="store_true",
            default=False,
            help="throw an error if configuration is bad, instead of assuming default",
        )
        parser.add_argument(
            "--stack-trace",
            dest="show_stack_trace",
            action="store_true",
            default=False,
            help="if an error occurs, print out the stack trace for debug purposes",
        )
        parser.add_argument(
            "--log-level",
            dest="log_level",
            action="store",
            help="minimum level required to log messages",
            type=PyMarkdownLint.log_level_type,
            choices=list(PyMarkdownLint.available_log_maps.keys()),
        )
        parser.add_argument(
            "--log-file",
            dest="log_file",
            action="store",
            help="destination file for log messages",
        )

        subparsers = parser.add_subparsers(dest="primary_subparser")

        PluginManager.add_argparse_subparser(subparsers)

        new_sub_parser = subparsers.add_parser(
            "scan", help="scan the Markdown files in the specified paths"
        )
        new_sub_parser.add_argument(
            "-l",
            "--list-files",
            dest="list_files",
            action="store_true",
            default=False,
            help="list the markdown files found and exit",
        )
        new_sub_parser.add_argument(
            "-r",
            "--recurse",
            dest="recurse_directories",
            action="store_true",
            default=False,
            help="recursively scan directories",
        )
        new_sub_parser.add_argument(
            "paths",
            metavar="path",
            type=str,
            nargs="+",
            help="one or more paths to scan for eligible Markdown files",
        )

        subparsers.add_parser("version", help="version of the application")

        parse_arguments = parser.parse_args()

        if not parse_arguments.primary_subparser:
            parser.print_help()
            sys.exit(2)
        elif parse_arguments.primary_subparser == "version":
            print(f"{self.__version_number}")
            sys.exit(0)
        return parse_arguments

    @classmethod
    def __is_file_eligible_to_scan(cls, path_to_test):
        """
        Determine if the presented path is one that we want to scan.
        """
        return path_to_test.endswith(".md")

    # pylint: disable=broad-except
    def __scan_file(self, args, next_file):
        """
        Scan a given file and call the plugin manager for any significant events.
        """

        POGGER.info("Scanning file '$'.", next_file)
        context = self.__plugins.starting_new_file(next_file)

        POGGER.info("Scanning file '$' token-by-token.", next_file)
        source_provider = FileSourceProvider(next_file)
        if args.x_test_scan_fault:
            source_provider = None
        actual_tokens = self.__tokenizer.transform_from_provider(source_provider)

        if actual_tokens and actual_tokens[-1].is_pragma:
            self.__plugins.compile_pragmas(next_file, actual_tokens[-1])
            actual_tokens = actual_tokens[:-1]

        POGGER.info("Scanning file '$' tokens.", next_file)
        for next_token in actual_tokens:
            POGGER.info("Processing token: $", next_token)
            self.__plugins.next_token(context, next_token)

        POGGER.info("Scanning file '$' line-by-line.", next_file)
        source_provider = FileSourceProvider(next_file)
        line_number = 1
        next_line = source_provider.get_next_line()
        while next_line is not None:
            POGGER.info("Processing line $: $", line_number, next_line)
            self.__plugins.next_line(context, line_number, next_line)
            line_number += 1
            next_line = source_provider.get_next_line()

        POGGER.info("Completed scanning file '$'.", next_file)
        self.__plugins.completed_file(context, line_number)

    # pylint: enable=broad-except

    def __process_next_path(self, next_path, files_to_parse, recurse_directories):

        did_find_any = False
        POGGER.info("Determining files to scan for path '$'.", next_path)
        if not os.path.exists(next_path):
            print(
                f"Provided path '{next_path}' does not exist.",
                file=sys.stderr,
            )
            POGGER.debug("Provided path '$' does not exist.", next_path)
        elif os.path.isdir(next_path):
            POGGER.debug(
                "Provided path '$' is a directory. Walking directory.", next_path
            )
            did_find_any = True
            normalized_next_path = next_path.replace("\\", "/")
            for root, _, files in os.walk(next_path):
                normalized_root = root.replace("\\", "/")
                if not recurse_directories and normalized_root != normalized_next_path:
                    continue
                normalized_root = (
                    normalized_root[0:-1]
                    if normalized_root.endswith("/")
                    else normalized_root
                )
                for file in files:
                    rooted_file_path = f"{normalized_root}/{file}"
                    if self.__is_file_eligible_to_scan(rooted_file_path):
                        files_to_parse.add(rooted_file_path)
        else:
            if self.__is_file_eligible_to_scan(next_path):
                POGGER.debug(
                    "Provided path '$' is a valid file. Adding.",
                    next_path,
                )
                files_to_parse.add(next_path)
                did_find_any = True
            else:
                POGGER.debug(
                    "Provided path '$' is not a valid file. Skipping.",
                    next_path,
                )
                print(
                    f"Provided file path '{next_path}' is not a valid file. Skipping.",
                    file=sys.stderr,
                )
        return did_find_any

    def __determine_files_to_scan(self, eligible_paths, recurse_directories):

        did_error_scanning_files = False
        files_to_parse = set()
        for next_path in eligible_paths:
            if "*" in next_path or "?" in next_path:
                globbed_paths = glob.glob(next_path)
                if not globbed_paths:
                    print(
                        f"Provided glob path '{next_path}' did not match any files.",
                        file=sys.stderr,
                    )
                    did_error_scanning_files = True
                    break
                for next_globbed_path in globbed_paths:
                    next_globbed_path = next_globbed_path.replace("\\", "/")
                    self.__process_next_path(
                        next_globbed_path, files_to_parse, recurse_directories
                    )
            else:
                if not self.__process_next_path(
                    next_path, files_to_parse, recurse_directories
                ):
                    did_error_scanning_files = True
                    break

        files_to_parse = list(files_to_parse)
        files_to_parse.sort()

        POGGER.info("Number of files found: $", len(files_to_parse))
        return files_to_parse, did_error_scanning_files

    @classmethod
    def __handle_list_files(cls, files_to_scan):

        if files_to_scan:
            print("\n".join(files_to_scan))
            return 0
        print("No matching files found.", file=sys.stderr)
        return 1

    def __apply_configuration_to_plugins(self):

        try:
            self.__plugins.apply_configuration(self.__properties)
        except BadPluginError as this_exception:
            formatted_error = f"{str(type(this_exception).__name__)} encountered while configuring plugins:\n{str(this_exception)}"
            self.__handle_error(formatted_error, this_exception)

    def __initialize_parser(self, args):

        resource_path = None
        if args.x_test_init_fault:
            resource_path = "fredo"

        try:
            self.__tokenizer = TokenizedMarkdown(resource_path)
            self.__tokenizer.apply_configuration(self.__properties)
        except ValueError as this_exception:
            formatted_error = "Configuration Error: " + str(this_exception)
            self.__handle_error(formatted_error, this_exception)
        except BadTokenizationError as this_exception:
            formatted_error = f"{str(type(this_exception).__name__)} encountered while initializing tokenizer:\n{str(this_exception)}"
            self.__handle_error(formatted_error, this_exception)

    def __initialize_plugin_manager(self, args, plugin_dir):
        """
        Make sure all plugins are ready before being initialized.
        """

        self.__plugins = PluginManager()
        try:
            self.__plugins.initialize(
                plugin_dir,
                args.add_plugin,
                args.enable_rules,
                args.disable_rules,
                self.__properties,
                self.__show_stack_trace,
            )
        except BadPluginError as this_exception:
            formatted_error = f"BadPluginError encountered while loading plugins:\n{str(this_exception)}"
            self.__handle_error(formatted_error, this_exception)

    def __handle_error(self, formatted_error, thrown_error):

        show_error = self.__show_stack_trace or not isinstance(thrown_error, ValueError)
        LOGGER.warning(formatted_error, exc_info=show_error)

        print(f"\n\n{formatted_error}", file=sys.stderr)
        if self.__show_stack_trace:
            traceback.print_exc(file=sys.stderr)
        sys.exit(1)

    def __handle_scan_error(self, next_file, this_exception):

        formatted_error = f"{str(type(this_exception).__name__)} encountered while scanning '{next_file}':\n{str(this_exception)}"
        self.__handle_error(formatted_error, this_exception)

    def __set_initial_state(self, args):

        self.__show_stack_trace = args.show_stack_trace
        base_logger = logging.getLogger()
        base_logger.setLevel(
            logging.DEBUG if self.__show_stack_trace else logging.WARNING
        )

        if args.configuration_file:
            LOGGER.debug("Loading configuration file: %s", args.configuration_file)
            ApplicationPropertiesJsonLoader.load_and_set(
                self.__properties, args.configuration_file, self.__handle_error
            )
        if args.set_configuration:
            self.__properties.set_manual_property(args.set_configuration)

    def __initialize_strict_mode(self, args):
        effective_strict_configuration = args.strict_configuration
        if not effective_strict_configuration:
            effective_strict_configuration = self.__properties.get_boolean_property(
                "mode.strict-config", strict_mode=True
            )

        if effective_strict_configuration:
            self.__properties.enable_strict_mode()

    def __initialize_logging(self, args):

        new_handler = None

        effective_log_file = args.log_file
        if effective_log_file is None:
            effective_log_file = self.__properties.get_string_property("log.file")

        if effective_log_file:
            new_handler = logging.FileHandler(effective_log_file)
            logging.getLogger().addHandler(new_handler)
        else:
            temp_log_level = (
                logging.DEBUG if self.__show_stack_trace else logging.CRITICAL
            )
            logging.basicConfig(stream=sys.stdout, level=temp_log_level)

        effective_log_level = args.log_level if args.log_level else None
        if effective_log_level is None:
            effective_log_level = self.__properties.get_string_property(
                "log.level", valid_value_fn=PyMarkdownLint.log_level_type
            )
        if effective_log_level is None:
            effective_log_level = self.default_log_level

        log_level_to_enact = PyMarkdownLint.available_log_maps[effective_log_level]

        logging.getLogger().setLevel(log_level_to_enact)
        ParserLogger.sync_on_next_call()
        return new_handler

    def __initialize_plugins(self, args):
        try:
            plugin_dir = os.path.dirname(os.path.realpath(__file__))
            plugin_dir = os.path.join(plugin_dir, "plugins")
            self.__initialize_plugin_manager(args, plugin_dir)
            self.__apply_configuration_to_plugins()
        except ValueError as this_exception:
            formatted_error = f"{str(type(this_exception).__name__)} encountered while initializing plugins:\n{str(this_exception)}"
            self.__handle_error(formatted_error, this_exception)

    def main(self):
        """
        Main entrance point.
        """
        args = self.__parse_arguments()
        self.__set_initial_state(args)

        new_handler = None
        total_error_count = 0
        try:
            self.__initialize_strict_mode(args)
            new_handler = self.__initialize_logging(args)

            if args.primary_subparser == PluginManager.argparse_subparser_name():
                self.__initialize_plugins(args)
                return_code = self.__plugins.handle_argparse_subparser(args)
                sys.exit(return_code)

            POGGER.info("Determining files to scan.")
            files_to_scan, did_error_scanning_files = self.__determine_files_to_scan(
                args.paths, args.recurse_directories
            )
            if did_error_scanning_files:
                total_error_count = 1
            else:
                self.__initialize_plugins(args)
                self.__initialize_parser(args)

                if args.list_files:
                    POGGER.info(
                        "Sending list of files that would have been scanned to stdout."
                    )
                    return_code = self.__handle_list_files(files_to_scan)
                    sys.exit(return_code)

                for next_file in files_to_scan:
                    try:
                        self.__scan_file(args, next_file)
                    except BadPluginError as this_exception:
                        self.__handle_scan_error(next_file, this_exception)
                    except BadTokenizationError as this_exception:
                        self.__handle_scan_error(next_file, this_exception)
        except ValueError as this_exception:
            formatted_error = f"Configuration Error: {this_exception}"
            self.__handle_error(formatted_error, this_exception)
        finally:
            if new_handler:
                new_handler.close()

        # TODO self.__plugins.number_of_pragma_failures
        if self.__plugins.number_of_scan_failures or total_error_count:
            sys.exit(1)


if __name__ == "__main__":
    PyMarkdownLint().main()



"""
Module to provide tests related to the MD002 rule.
"""
import os
from test.markdown_scanner import MarkdownScanner

import pytest

from .utils import write_temporary_configuration


@pytest.mark.rules
def test_md002_all_samples():
    """
    Test to make sure we get the expected behavior after scanning the files in the
    test/resources/rules/md002 directory.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--disable-rules",
        "MD003",
        "-e",
        "MD002",
        "scan",
        "test/resources/rules/md002",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md002/improper_atx_heading_start.md:1:1: "
        + "MD002: First heading of the document should be a top level heading. "
        + "[Expected: h1; Actual: h2] (first-heading-h1,first-header-h1)\n"
        + "test/resources/rules/md002/improper_setext_heading_start.md:2:1: "
        + "MD002: First heading of the document should be a top level heading. "
        + "[Expected: h1; Actual: h2] (first-heading-h1,first-header-h1)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md002_good_proper_atx_heading_start():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md002 directory that starts with a h1 atx heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "-e",
        "MD002",
        "scan",
        "test/resources/rules/md002/proper_atx_heading_start.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md002_bad_proper_atx_heading_start_with_alternate_configuration():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md002 directory that starts with a h1 atx heading.  The modified
    configuration will cause the file to be parsed with reported issues.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {"plugins": {"md002": {"level": 2}}}
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "-e",
            "MD002",
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md002/proper_atx_heading_start.md",
        ]

        expected_return_code = 1
        expected_output = (
            "test/resources/rules/md002/proper_atx_heading_start.md:1:1: "
            + "MD002: First heading of the document should be a top level heading. "
            + "[Expected: h2; Actual: h1] (first-heading-h1,first-header-h1)\n"
        )
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)


@pytest.mark.rules
def test_md002_good_proper_setext_heading_start():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md002 directory starting with a setext heading of ====== to
    create a h1 heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "-e",
        "MD002",
        "scan",
        "test/resources/rules/md002/proper_setext_heading_start.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md002_bad_proper_setext_heading_start_with_alternate_configuration():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md002 directory starting with a setext heading of ====== to
    create a h1 heading.  The modified  configuration will cause the file to be parsed
    with reported issues.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {"plugins": {"md002": {"level": 2}}}
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "-e",
            "MD002",
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md002/proper_setext_heading_start.md",
        ]

        expected_return_code = 1
        expected_output = (
            "test/resources/rules/md002/proper_setext_heading_start.md:2:1: "
            + "MD002: First heading of the document should be a top level heading. "
            + "[Expected: h2; Actual: h1] (first-heading-h1,first-header-h1)\n"
        )
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)


@pytest.mark.rules
def test_md002_bad_improper_atx_heading_start():
    """
    Test to make sure we get the expected behavior after scanning a bad file from the
    test/resources/rules/md002 directory that starts with a non-h1 heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "-e",
        "MD002",
        "scan",
        "test/resources/rules/md002/improper_atx_heading_start.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md002/improper_atx_heading_start.md:1:1: "
        + "MD002: First heading of the document should be a top level heading. "
        + "[Expected: h1; Actual: h2] (first-heading-h1,first-header-h1)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md002_good_improper_atx_heading_start_with_alternate_configuration():
    """
    Test to make sure we get the expected behavior after scanning a bad file from the
    test/resources/rules/md002 directory that starts with a non-h1 heading.  The modified
    configuration will allow the file to be parsed properly.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {"plugins": {"md002": {"level": 2}}}
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "-e",
            "MD002",
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md002/improper_atx_heading_start.md",
        ]

        expected_return_code = 0
        expected_output = ""
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)


@pytest.mark.rules
def test_md002_bad_improper_setext_heading_start():
    """
    Test to make sure we get the expected behavior after scanning a bad file from the
    test/resources/rules/md002 directory starting with a setext heading of ----- to
    create a h2 heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--disable-rules",
        "MD003",
        "-e",
        "MD002",
        "scan",
        "test/resources/rules/md002/improper_setext_heading_start.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md002/improper_setext_heading_start.md:2:1: "
        + "MD002: First heading of the document should be a top level heading. "
        + "[Expected: h1; Actual: h2] (first-heading-h1,first-header-h1)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md002_good_improper_setext_heading_start_with_alternate_configuration():
    """
    Test to make sure we get the expected behavior after scanning a bad file from the
    test/resources/rules/md002 directory starting with a setext heading of ----- to
    create a h2 heading.   The modified configuration will allow the file to be parsed
    properly.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {"plugins": {"md002": {"level": 2}}}
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "--disable-rules",
            "MD003",
            "-e",
            "MD002",
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md002/improper_setext_heading_start.md",
        ]

        expected_return_code = 0
        expected_output = ""
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)



"""
https://github.github.com/gfm/#lists
"""
import pytest

from .utils import act_and_assert


# pylint: disable=too-many-lines
@pytest.mark.gfm
def test_list_items_281():
    """
    Test case 281:  (part 1) Changing the bullet or ordered list delimiter starts a new list:
    """

    # Arrange
    source_markdown = """- foo
- bar
+ baz"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[li(2,1):2::]",
        "[para(2,3):]",
        "[text(2,3):bar:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[ulist(3,1):+::2:]",
        "[para(3,3):]",
        "[text(3,3):baz:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>foo</li>
<li>bar</li>
</ul>
<ul>
<li>baz</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_281a():
    """
    Test case 281a:  variation on 281 with second item indented
    """

    # Arrange
    source_markdown = """* foo
  * bar
* baz"""
    expected_tokens = [
        "[ulist(1,1):*::2:]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[ulist(2,3):*::4:  ]",
        "[para(2,5):]",
        "[text(2,5):bar:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[li(3,1):2::]",
        "[para(3,3):]",
        "[text(3,3):baz:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>foo
<ul>
<li>bar</li>
</ul>
</li>
<li>baz</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_281b():
    """
    Test case 281b:  variation on 281 with second item indented and third with different list start
    """

    # Arrange
    source_markdown = """* foo
  * bar
+ baz"""
    expected_tokens = [
        "[ulist(1,1):*::2:]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[ulist(2,3):*::4:  ]",
        "[para(2,5):]",
        "[text(2,5):bar:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-ulist:::True]",
        "[ulist(3,1):+::2:]",
        "[para(3,3):]",
        "[text(3,3):baz:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>foo
<ul>
<li>bar</li>
</ul>
</li>
</ul>
<ul>
<li>baz</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_281c():
    """
    Test case 281c:  variation on 281b with third item indented, and a following list item for the parent
    """

    # Arrange
    source_markdown = """* foo
  * bar
  + baz
* boffo"""
    expected_tokens = [
        "[ulist(1,1):*::2:]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[ulist(2,3):*::4:  ]",
        "[para(2,5):]",
        "[text(2,5):bar:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[ulist(3,3):+::4:  ]",
        "[para(3,5):]",
        "[text(3,5):baz:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[li(4,1):2::]",
        "[para(4,3):]",
        "[text(4,3):boffo:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>foo
<ul>
<li>bar</li>
</ul>
<ul>
<li>baz</li>
</ul>
</li>
<li>boffo</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_281d():
    """
    Test case 281d:  variation on 281
    """

    # Arrange
    source_markdown = """* foo
  * bar
    * boffo
  + baz"""
    expected_tokens = [
        "[ulist(1,1):*::2:]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[ulist(2,3):*::4:  ]",
        "[para(2,5):]",
        "[text(2,5):bar:]",
        "[end-para:::True]",
        "[ulist(3,5):*::6:    ]",
        "[para(3,7):]",
        "[text(3,7):boffo:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-ulist:::True]",
        "[ulist(4,3):+::4:  ]",
        "[para(4,5):]",
        "[text(4,5):baz:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>foo
<ul>
<li>bar
<ul>
<li>boffo</li>
</ul>
</li>
</ul>
<ul>
<li>baz</li>
</ul>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_281e():
    """
    Test case 281e:  variation on 281
    """

    # Arrange
    source_markdown = """* foo
  * bar
    * boffo
+ baz"""
    expected_tokens = [
        "[ulist(1,1):*::2:]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[ulist(2,3):*::4:  ]",
        "[para(2,5):]",
        "[text(2,5):bar:]",
        "[end-para:::True]",
        "[ulist(3,5):*::6:    ]",
        "[para(3,7):]",
        "[text(3,7):boffo:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-ulist:::True]",
        "[end-ulist:::True]",
        "[ulist(4,1):+::2:]",
        "[para(4,3):]",
        "[text(4,3):baz:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>foo
<ul>
<li>bar
<ul>
<li>boffo</li>
</ul>
</li>
</ul>
</li>
</ul>
<ul>
<li>baz</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_282():
    """
    Test case 282:  (part 2) Changing the bullet or ordered list delimiter starts a new list:
    """

    # Arrange
    source_markdown = """1. foo
2. bar
3) baz"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):foo:]",
        "[end-para:::True]",
        "[li(2,1):3::2]",
        "[para(2,4):]",
        "[text(2,4):bar:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[olist(3,1):):3:3:]",
        "[para(3,4):]",
        "[text(3,4):baz:]",
        "[end-para:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>foo</li>
<li>bar</li>
</ol>
<ol start="3">
<li>baz</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_282a():
    """
    Test case 282a:  variation
    """

    # Arrange
    source_markdown = """1. foo
2. bar
 3) baz"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):foo:]",
        "[end-para:::True]",
        "[li(2,1):3::2]",
        "[para(2,4):]",
        "[text(2,4):bar:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[olist(3,2):):3:4: ]",
        "[para(3,5):]",
        "[text(3,5):baz:]",
        "[end-para:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>foo</li>
<li>bar</li>
</ol>
<ol start="3">
<li>baz</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_282b():
    """
    Test case 282b:  variation
    """

    # Arrange
    source_markdown = """1. foo
2. bar
  3) baz"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):foo:]",
        "[end-para:::True]",
        "[li(2,1):3::2]",
        "[para(2,4):]",
        "[text(2,4):bar:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[olist(3,3):):3:5:  ]",
        "[para(3,6):]",
        "[text(3,6):baz:]",
        "[end-para:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>foo</li>
<li>bar</li>
</ol>
<ol start="3">
<li>baz</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_282c():
    """
    Test case 282c:  variation
    """

    # Arrange
    source_markdown = """1. foo
2. bar
   3) baz"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   ]",
        "[para(1,4):]",
        "[text(1,4):foo:]",
        "[end-para:::True]",
        "[li(2,1):3::2]",
        "[para(2,4):\n]",
        "[text(2,4):bar\n3) baz::\n]",
        "[end-para:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>foo</li>
<li>bar
3) baz</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_282cu():
    """
    Test case 282c:  variation
    """

    # Arrange
    source_markdown = """- foo
- bar
  * baz"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[li(2,1):2::]",
        "[para(2,3):]",
        "[text(2,3):bar:]",
        "[end-para:::True]",
        "[ulist(3,3):*::4:  ]",
        "[para(3,5):]",
        "[text(3,5):baz:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>foo</li>
<li>bar
<ul>
<li>baz</li>
</ul>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_282d():
    """
    Test case 282d:  variation
    """

    # Arrange
    source_markdown = """1. foo
2. bar

   3) baz"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):foo:]",
        "[end-para:::True]",
        "[li(2,1):3::2]",
        "[para(2,4):]",
        "[text(2,4):bar:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[olist(4,4):):3:6:   ]",
        "[para(4,7):]",
        "[text(4,7):baz:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>
<p>foo</p>
</li>
<li>
<p>bar</p>
<ol start="3">
<li>baz</li>
</ol>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_283():
    """
    Test case 283:  In CommonMark, a list can interrupt a paragraph. That is, no blank line is needed to separate a paragraph from a following list:
    """

    # Arrange
    source_markdown = """Foo
- bar
- baz"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):Foo:]",
        "[end-para:::True]",
        "[ulist(2,1):-::2:]",
        "[para(2,3):]",
        "[text(2,3):bar:]",
        "[end-para:::True]",
        "[li(3,1):2::]",
        "[para(3,3):]",
        "[text(3,3):baz:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<p>Foo</p>
<ul>
<li>bar</li>
<li>baz</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_284():
    """
    Test case 284:  In order to solve of unwanted lists in paragraphs with hard-wrapped numerals, we allow only lists starting with 1 to interrupt paragraphs. Thus,
    """

    # Arrange
    source_markdown = """The number of windows in my house is
14.  The number of doors is 6."""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):The number of windows in my house is\n14.  The number of doors is 6.::\n]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>The number of windows in my house is
14.  The number of doors is 6.</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_285():
    """
    Test case 285:  We may still get an unintended result in cases like
    """

    # Arrange
    source_markdown = """The number of windows in my house is
1.  The number of doors is 6."""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):The number of windows in my house is:]",
        "[end-para:::True]",
        "[olist(2,1):.:1:4:]",
        "[para(2,5):]",
        "[text(2,5):The number of doors is 6.:]",
        "[end-para:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<p>The number of windows in my house is</p>
<ol>
<li>The number of doors is 6.</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_286():
    """
    Test case 286:  (part 1) There can be any number of blank lines between items:
    """

    # Arrange
    source_markdown = """- foo

- bar


- baz"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[li(3,1):2::]",
        "[para(3,3):]",
        "[text(3,3):bar:]",
        "[end-para:::True]",
        "[BLANK(4,1):]",
        "[BLANK(5,1):]",
        "[li(6,1):2::]",
        "[para(6,3):]",
        "[text(6,3):baz:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>
<p>foo</p>
</li>
<li>
<p>bar</p>
</li>
<li>
<p>baz</p>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_287():
    """
    Test case 287:  (part 2) There can be any number of blank lines between items:
    """

    # Arrange
    source_markdown = """- foo
  - bar
    - baz


      bim"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[ulist(2,3):-::4:  ]",
        "[para(2,5):]",
        "[text(2,5):bar:]",
        "[end-para:::True]",
        "[ulist(3,5):-::6:    :      ]",
        "[para(3,7):]",
        "[text(3,7):baz:]",
        "[end-para:::True]",
        "[BLANK(4,1):]",
        "[BLANK(5,1):]",
        "[para(6,7):]",
        "[text(6,7):bim:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-ulist:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>foo
<ul>
<li>bar
<ul>
<li>
<p>baz</p>
<p>bim</p>
</li>
</ul>
</li>
</ul>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_288():
    """
    Test case 288:  (part 1) To separate consecutive lists of the same type, or to separate a list from an indented code block that would otherwise be parsed as a subparagraph of the final list item, you can insert a blank HTML comment:
    """

    # Arrange
    source_markdown = """- foo
- bar

<!-- -->

- baz
- bim"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[li(2,1):2::]",
        "[para(2,3):]",
        "[text(2,3):bar:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[end-ulist:::True]",
        "[html-block(4,1)]",
        "[text(4,1):<!-- -->:]",
        "[end-html-block:::False]",
        "[BLANK(5,1):]",
        "[ulist(6,1):-::2:]",
        "[para(6,3):]",
        "[text(6,3):baz:]",
        "[end-para:::True]",
        "[li(7,1):2::]",
        "[para(7,3):]",
        "[text(7,3):bim:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>foo</li>
<li>bar</li>
</ul>
<!-- -->
<ul>
<li>baz</li>
<li>bim</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_289():
    """
    Test case 289:  (part 2) To separate consecutive lists of the same type, or to separate a list from an indented code block that would otherwise be parsed as a subparagraph of the final list item, you can insert a blank HTML comment:
    """

    # Arrange
    source_markdown = """-   foo

    notcode

-   foo

<!-- -->

    code"""
    expected_tokens = [
        "[ulist(1,1):-::4::    ]",
        "[para(1,5):]",
        "[text(1,5):foo:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[para(3,5):]",
        "[text(3,5):notcode:]",
        "[end-para:::True]",
        "[BLANK(4,1):]",
        "[li(5,1):4::]",
        "[para(5,5):]",
        "[text(5,5):foo:]",
        "[end-para:::True]",
        "[BLANK(6,1):]",
        "[end-ulist:::True]",
        "[html-block(7,1)]",
        "[text(7,1):<!-- -->:]",
        "[end-html-block:::False]",
        "[BLANK(8,1):]",
        "[icode-block(9,5):    :]",
        "[text(9,5):code:]",
        "[end-icode-block:::True]",
    ]
    expected_gfm = """<ul>
<li>
<p>foo</p>
<p>notcode</p>
</li>
<li>
<p>foo</p>
</li>
</ul>
<!-- -->
<pre><code>code
</code></pre>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_290():
    """
    Test case 290:  (part 1) List items need not be indented to the same level. The following list items will be treated as items at the same list level, since none is indented enough to belong to the previous list item:
    """

    # Arrange
    source_markdown = """- a
 - b
  - c
   - d
  - e
 - f
- g"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[li(2,2):3: :]",
        "[para(2,4):]",
        "[text(2,4):b:]",
        "[end-para:::True]",
        "[li(3,3):4:  :]",
        "[para(3,5):]",
        "[text(3,5):c:]",
        "[end-para:::True]",
        "[li(4,4):5:   :]",
        "[para(4,6):]",
        "[text(4,6):d:]",
        "[end-para:::True]",
        "[li(5,3):4:  :]",
        "[para(5,5):]",
        "[text(5,5):e:]",
        "[end-para:::True]",
        "[li(6,2):3: :]",
        "[para(6,4):]",
        "[text(6,4):f:]",
        "[end-para:::True]",
        "[li(7,1):2::]",
        "[para(7,3):]",
        "[text(7,3):g:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>a</li>
<li>b</li>
<li>c</li>
<li>d</li>
<li>e</li>
<li>f</li>
<li>g</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_291():
    """
    Test case 291:  (part 2) List items need not be indented to the same level. The following list items will be treated as items at the same list level, since none is indented enough to belong to the previous list item:
    """

    # Arrange
    source_markdown = """1. a

  2. b

   3. c"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):a:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[li(3,3):5:  :2]",
        "[para(3,6):]",
        "[text(3,6):b:]",
        "[end-para:::True]",
        "[BLANK(4,1):]",
        "[li(5,4):6:   :3]",
        "[para(5,7):]",
        "[text(5,7):c:]",
        "[end-para:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>
<p>a</p>
</li>
<li>
<p>b</p>
</li>
<li>
<p>c</p>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_292x():
    """
    Test case 292x:  Note, however, that list items may not be indented more than three spaces. Here - e is treated as a paragraph continuation line, because it is indented more than three spaces:
    """

    # Arrange
    source_markdown = """- a
 - b
  - c
   - d
    - e"""
    expected_tokens = [
        "[ulist(1,1):-::2::  ]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[li(2,2):3: :]",
        "[para(2,4):]",
        "[text(2,4):b:]",
        "[end-para:::True]",
        "[li(3,3):4:  :]",
        "[para(3,5):]",
        "[text(3,5):c:]",
        "[end-para:::True]",
        "[li(4,4):5:   :]",
        "[para(4,6):\n\x04]",
        "[text(4,6):d\n- e::\n]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>a</li>
<li>b</li>
<li>c</li>
<li>d
- e</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_292xa():
    """
    Test case 292xa:  variation, indent by 2 instead of 1
    """

    # Arrange
    source_markdown = """- a
  - b
    - c
      - d
        - e"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[ulist(2,3):-::4:  ]",
        "[para(2,5):]",
        "[text(2,5):b:]",
        "[end-para:::True]",
        "[ulist(3,5):-::6:    ]",
        "[para(3,7):]",
        "[text(3,7):c:]",
        "[end-para:::True]",
        "[ulist(4,7):-::8:      ]",
        "[para(4,9):]",
        "[text(4,9):d:]",
        "[end-para:::True]",
        "[ulist(5,9):-::10:        ]",
        "[para(5,11):]",
        "[text(5,11):e:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-ulist:::True]",
        "[end-ulist:::True]",
        "[end-ulist:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>a
<ul>
<li>b
<ul>
<li>c
<ul>
<li>d
<ul>
<li>e</li>
</ul>
</li>
</ul>
</li>
</ul>
</li>
</ul>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_292a():
    """
    Test case 292a:  Variation on 292
    """

    # Arrange
    source_markdown = """- a
 - b
  - c
- d"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[li(2,2):3: :]",
        "[para(2,4):]",
        "[text(2,4):b:]",
        "[end-para:::True]",
        "[li(3,3):4:  :]",
        "[para(3,5):]",
        "[text(3,5):c:]",
        "[end-para:::True]",
        "[li(4,1):2::]",
        "[para(4,3):]",
        "[text(4,3):d:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>a</li>
<li>b</li>
<li>c</li>
<li>d</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_292b():
    """
    Test case 292b:  Variation on 292
    """

    # Arrange
    source_markdown = """- a
 - b
  - c
   - d
- e"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[li(2,2):3: :]",
        "[para(2,4):]",
        "[text(2,4):b:]",
        "[end-para:::True]",
        "[li(3,3):4:  :]",
        "[para(3,5):]",
        "[text(3,5):c:]",
        "[end-para:::True]",
        "[li(4,4):5:   :]",
        "[para(4,6):]",
        "[text(4,6):d:]",
        "[end-para:::True]",
        "[li(5,1):2::]",
        "[para(5,3):]",
        "[text(5,3):e:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>a</li>
<li>b</li>
<li>c</li>
<li>d</li>
<li>e</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_292c():
    """
    Test case 292c:  variations
    """

    # Arrange
    source_markdown = """1. a
 1. b
  1. c
   1. d
    1. e"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   ]",
        "[para(1,4):]",
        "[text(1,4):a:]",
        "[end-para:::True]",
        "[li(2,2):4: :1]",
        "[para(2,5):]",
        "[text(2,5):b:]",
        "[end-para:::True]",
        "[li(3,3):5:  :1]",
        "[para(3,6):]",
        "[text(3,6):c:]",
        "[end-para:::True]",
        "[li(4,4):6:   :1]",
        "[para(4,7):\n\x04\x04]",
        "[text(4,7):d\n1. e::\n]",
        "[end-para:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>a</li>
<li>b</li>
<li>c</li>
<li>d
1. e</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_292d():
    """
    Test case 292a:  Variation on 292
    """

    # Arrange
    source_markdown = """1. a
 1. b
  1. c
1. d"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):a:]",
        "[end-para:::True]",
        "[li(2,2):4: :1]",
        "[para(2,5):]",
        "[text(2,5):b:]",
        "[end-para:::True]",
        "[li(3,3):5:  :1]",
        "[para(3,6):]",
        "[text(3,6):c:]",
        "[end-para:::True]",
        "[li(4,1):3::1]",
        "[para(4,4):]",
        "[text(4,4):d:]",
        "[end-para:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>a</li>
<li>b</li>
<li>c</li>
<li>d</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_292e():
    """
    Test case 292b:  Variation on 292
    """

    # Arrange
    source_markdown = """1. a
 1. b
  1. c
   1. d
1. e"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):a:]",
        "[end-para:::True]",
        "[li(2,2):4: :1]",
        "[para(2,5):]",
        "[text(2,5):b:]",
        "[end-para:::True]",
        "[li(3,3):5:  :1]",
        "[para(3,6):]",
        "[text(3,6):c:]",
        "[end-para:::True]",
        "[li(4,4):6:   :1]",
        "[para(4,7):]",
        "[text(4,7):d:]",
        "[end-para:::True]",
        "[li(5,1):3::1]",
        "[para(5,4):]",
        "[text(5,4):e:]",
        "[end-para:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>a</li>
<li>b</li>
<li>c</li>
<li>d</li>
<li>e</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_293():
    """
    Test case 293:  And here, 3. c is treated as in indented code block, because it is indented four spaces and preceded by a blank line.
    """

    # Arrange
    source_markdown = """1. a

  2. b

    3. c"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):a:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[li(3,3):5:  :2]",
        "[para(3,6):]",
        "[text(3,6):b:]",
        "[end-para:::True]",
        "[BLANK(4,1):]",
        "[end-olist:::True]",
        "[icode-block(5,5):    :]",
        "[text(5,5):3. c:]",
        "[end-icode-block:::True]",
    ]
    expected_gfm = """<ol>
<li>
<p>a</p>
</li>
<li>
<p>b</p>
</li>
</ol>
<pre><code>3. c
</code></pre>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_293a():
    """
    Test case 293a:  variation on 293
    """

    # Arrange
    source_markdown = """1. a
  1. b
    1. c"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   ]",
        "[para(1,4):]",
        "[text(1,4):a:]",
        "[end-para:::True]",
        "[li(2,3):5:  :1]",
        "[para(2,6):\n\x04]",
        "[text(2,6):b\n1. c::\n]",
        "[end-para:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>a</li>
<li>b
1. c</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_293b():
    """
    Test case 293b:  variation on 293
    """

    # Arrange
    source_markdown = """1. a
   1. b
1. c"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):a:]",
        "[end-para:::True]",
        "[olist(2,4):.:1:6:   ]",
        "[para(2,7):]",
        "[text(2,7):b:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[li(3,1):3::1]",
        "[para(3,4):]",
        "[text(3,4):c:]",
        "[end-para:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>a
<ol>
<li>b</li>
</ol>
</li>
<li>c</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_293c():
    """
    Test case 293c:  variation on 293
    """

    # Arrange
    source_markdown = """1. a
   1. b
1) c"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):a:]",
        "[end-para:::True]",
        "[olist(2,4):.:1:6:   ]",
        "[para(2,7):]",
        "[text(2,7):b:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[olist(3,1):):1:3:]",
        "[para(3,4):]",
        "[text(3,4):c:]",
        "[end-para:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>a
<ol>
<li>b</li>
</ol>
</li>
</ol>
<ol>
<li>c</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_293d():
    """
    Test case 293d:  variation on 293
    """

    # Arrange
    source_markdown = """1. a
   1. b
   1) c
1. d"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):a:]",
        "[end-para:::True]",
        "[olist(2,4):.:1:6:   ]",
        "[para(2,7):]",
        "[text(2,7):b:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[olist(3,4):):1:6:   ]",
        "[para(3,7):]",
        "[text(3,7):c:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[li(4,1):3::1]",
        "[para(4,4):]",
        "[text(4,4):d:]",
        "[end-para:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>a
<ol>
<li>b</li>
</ol>
<ol>
<li>c</li>
</ol>
</li>
<li>d</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_293e():
    """
    Test case 293e:  variation on 293
    """

    # Arrange
    source_markdown = """1. a
   1. b
      1. c
   1) d"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):a:]",
        "[end-para:::True]",
        "[olist(2,4):.:1:6:   ]",
        "[para(2,7):]",
        "[text(2,7):b:]",
        "[end-para:::True]",
        "[olist(3,7):.:1:9:      ]",
        "[para(3,10):]",
        "[text(3,10):c:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[olist(4,4):):1:6:   ]",
        "[para(4,7):]",
        "[text(4,7):d:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>a
<ol>
<li>b
<ol>
<li>c</li>
</ol>
</li>
</ol>
<ol>
<li>d</li>
</ol>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_293f():
    """
    Test case 293f:  variation on 293
    """

    # Arrange
    source_markdown = """1. a
   1. b
      1. c
1) d"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):a:]",
        "[end-para:::True]",
        "[olist(2,4):.:1:6:   ]",
        "[para(2,7):]",
        "[text(2,7):b:]",
        "[end-para:::True]",
        "[olist(3,7):.:1:9:      ]",
        "[para(3,10):]",
        "[text(3,10):c:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[olist(4,1):):1:3:]",
        "[para(4,4):]",
        "[text(4,4):d:]",
        "[end-para:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>a
<ol>
<li>b
<ol>
<li>c</li>
</ol>
</li>
</ol>
</li>
</ol>
<ol>
<li>d</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_294():
    """
    Test case 294:  This is a loose list, because there is a blank line between two of the list items:
    """

    # Arrange
    source_markdown = """- a
- b

- c"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[li(2,1):2::]",
        "[para(2,3):]",
        "[text(2,3):b:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[li(4,1):2::]",
        "[para(4,3):]",
        "[text(4,3):c:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>
<p>a</p>
</li>
<li>
<p>b</p>
</li>
<li>
<p>c</p>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_295():
    """
    Test case 295:  So is this, with a empty second item:
    """

    # Arrange
    source_markdown = """* a
*

* c"""
    expected_tokens = [
        "[ulist(1,1):*::2:]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[li(2,1):2::]",
        "[BLANK(2,2):]",
        "[BLANK(3,1):]",
        "[li(4,1):2::]",
        "[para(4,3):]",
        "[text(4,3):c:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>
<p>a</p>
</li>
<li></li>
<li>
<p>c</p>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_296():
    """
    Test case 296:  (part 1) These are loose lists, even though there is no space between the items, because one of the items directly contains two block-level elements with a blank line between them:
    """

    # Arrange
    source_markdown = """- a
- b

  c
- d"""
    expected_tokens = [
        "[ulist(1,1):-::2::  ]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[li(2,1):2::]",
        "[para(2,3):]",
        "[text(2,3):b:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[para(4,3):]",
        "[text(4,3):c:]",
        "[end-para:::True]",
        "[li(5,1):2::]",
        "[para(5,3):]",
        "[text(5,3):d:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>
<p>a</p>
</li>
<li>
<p>b</p>
<p>c</p>
</li>
<li>
<p>d</p>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_297():
    """
    Test case 297:  (part 2) These are loose lists, even though there is no space between the items, because one of the items directly contains two block-level elements with a blank line between them:
    """

    # Arrange
    source_markdown = """- a
- b

  [ref]: /url
- d"""
    expected_tokens = [
        "[ulist(1,1):-::2::  ]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[li(2,1):2::]",
        "[para(2,3):]",
        "[text(2,3):b:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[link-ref-def(4,3):True::ref:: :/url:::::]",
        "[li(5,1):2::]",
        "[para(5,3):]",
        "[text(5,3):d:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>
<p>a</p>
</li>
<li>
<p>b</p>
</li>
<li>
<p>d</p>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_297a():
    """
    Test case 297a:  variation
    """

    # Arrange
    source_markdown = """- a
- b

  # Heading
- d"""
    expected_tokens = [
        "[ulist(1,1):-::2::  ]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[li(2,1):2::]",
        "[para(2,3):]",
        "[text(2,3):b:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[atx(4,3):1:0:]",
        "[text(4,5):Heading: ]",
        "[end-atx::]",
        "[li(5,1):2::]",
        "[para(5,3):]",
        "[text(5,3):d:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>
<p>a</p>
</li>
<li>
<p>b</p>
<h1>Heading</h1>
</li>
<li>
<p>d</p>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_297b():
    """
    Test case 297b:  variation
    """

    # Arrange
    source_markdown = """- a
- b

  Heading
  -------
- d"""
    expected_tokens = [
        "[ulist(1,1):-::2::  \n  ]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[li(2,1):2::]",
        "[para(2,3):]",
        "[text(2,3):b:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[setext(5,3):-:7::(4,3)]",
        "[text(4,3):Heading:]",
        "[end-setext::]",
        "[li(6,1):2::]",
        "[para(6,3):]",
        "[text(6,3):d:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>
<p>a</p>
</li>
<li>
<p>b</p>
<h2>Heading</h2>
</li>
<li>
<p>d</p>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_297c():
    """
    Test case 297c:  variation
    """

    # Arrange
    source_markdown = """- a
- b

  <!-- script
  line
  -->
- d"""
    expected_tokens = [
        "[ulist(1,1):-::2::  \n  \n  ]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[li(2,1):2::]",
        "[para(2,3):]",
        "[text(2,3):b:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[html-block(4,3)]",
        "[text(4,3):<!-- script\nline\n-->:]",
        "[end-html-block:::False]",
        "[li(7,1):2::]",
        "[para(7,3):]",
        "[text(7,3):d:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>
<p>a</p>
</li>
<li>
<p>b</p>
<!-- script
line
-->
</li>
<li>
<p>d</p>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_297d():
    """
    Test case 297d:  variation
    """

    # Arrange
    source_markdown = """- a
- b

      script
- d"""
    expected_tokens = [
        "[ulist(1,1):-::2::  ]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[li(2,1):2::]",
        "[para(2,3):]",
        "[text(2,3):b:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[icode-block(4,7):    :]",
        "[text(4,7):script:]",
        "[end-icode-block:::True]",
        "[li(5,1):2::]",
        "[para(5,3):]",
        "[text(5,3):d:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>
<p>a</p>
</li>
<li>
<p>b</p>
<pre><code>script
</code></pre>
</li>
<li>
<p>d</p>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_297e():
    """
    Test case 297e:  variation
    """

    # Arrange
    source_markdown = """- a
- b

   ```script
   my script
   ```
- d"""
    expected_tokens = [
        "[ulist(1,1):-::2::  \n  \n  ]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[li(2,1):2::]",
        "[para(2,3):]",
        "[text(2,3):b:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[fcode-block(4,4):`:3:script:::: :]",
        "[text(5,3):my script:\a \a\x03\a]",
        "[end-fcode-block: :3:False]",
        "[li(7,1):2::]",
        "[para(7,3):]",
        "[text(7,3):d:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>
<p>a</p>
</li>
<li>
<p>b</p>
<pre><code class="language-script">my script
</code></pre>
</li>
<li>
<p>d</p>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_297f():
    """
    Test case 297f:  variation
    """

    # Arrange
    source_markdown = """- a
- b

   ```script
   my script
   ```
  <!-- script
  line
  -->
- d"""
    expected_tokens = [
        "[ulist(1,1):-::2::  \n  \n  \n  \n  \n  ]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[li(2,1):2::]",
        "[para(2,3):]",
        "[text(2,3):b:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[fcode-block(4,4):`:3:script:::: :]",
        "[text(5,3):my script:\a \a\x03\a]",
        "[end-fcode-block: :3:False]",
        "[html-block(7,3)]",
        "[text(7,3):<!-- script\nline\n-->:]",
        "[end-html-block:::False]",
        "[li(10,1):2::]",
        "[para(10,3):]",
        "[text(10,3):d:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>
<p>a</p>
</li>
<li>
<p>b</p>
<pre><code class="language-script">my script
</code></pre>
<!-- script
line
-->
</li>
<li>
<p>d</p>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_298():
    """
    Test case 298:  This is a tight list, because the blank lines are in a code block:
    """

    # Arrange
    source_markdown = """- a
- ```
  b


  ```
- c"""
    expected_tokens = [
        "[ulist(1,1):-::2::  \n  ]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[li(2,1):2::]",
        "[fcode-block(2,3):`:3::::::]",
        "[text(3,3):b\n\x03\n\x03:]",
        "[end-fcode-block::3:False]",
        "[li(7,1):2::]",
        "[para(7,3):]",
        "[text(7,3):c:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>a</li>
<li>
<pre><code>b


</code></pre>
</li>
<li>c</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_299():
    """
    Test case 299:  This is a tight list, because the blank line is between two paragraphs of a sublist. So the sublist is loose while the outer list is tight:
    """

    # Arrange
    source_markdown = """- a
  - b

    c
- d"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[ulist(2,3):-::4:  :    ]",
        "[para(2,5):]",
        "[text(2,5):b:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[para(4,5):]",
        "[text(4,5):c:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[li(5,1):2::]",
        "[para(5,3):]",
        "[text(5,3):d:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>a
<ul>
<li>
<p>b</p>
<p>c</p>
</li>
</ul>
</li>
<li>d</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_300():
    """
    Test case 300:  This is a tight list, because the blank line is inside the block quote:
    """

    # Arrange
    source_markdown = """* a
  > b
  >
* c"""
    expected_tokens = [
        "[ulist(1,1):*::2:]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > \n  >]",
        "[para(2,5):]",
        "[text(2,5):b:]",
        "[end-para:::True]",
        "[BLANK(3,4):]",
        "[end-block-quote:::True]",
        "[li(4,1):2::]",
        "[para(4,3):]",
        "[text(4,3):c:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>a
<blockquote>
<p>b</p>
</blockquote>
</li>
<li>c</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_300a():
    """
    Test case 300a:  variation
    """

    # Arrange
    source_markdown = """* a
  > b
  >
1) c"""
    expected_tokens = [
        "[ulist(1,1):*::2:]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > \n  >]",
        "[para(2,5):]",
        "[text(2,5):b:]",
        "[end-para:::True]",
        "[BLANK(3,4):]",
        "[end-block-quote:::True]",
        "[end-ulist:::True]",
        "[olist(4,1):):1:3:]",
        "[para(4,4):]",
        "[text(4,4):c:]",
        "[end-para:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ul>
<li>a
<blockquote>
<p>b</p>
</blockquote>
</li>
</ul>
<ol>
<li>c</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_300b():
    """
    Test case 300b:  variation
    """

    # Arrange
    source_markdown = """* a
  > b
  >
  * c"""
    expected_tokens = [
        "[ulist(1,1):*::2:]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > \n  >]",
        "[para(2,5):]",
        "[text(2,5):b:]",
        "[end-para:::True]",
        "[BLANK(3,4):]",
        "[end-block-quote:::True]",
        "[ulist(4,3):*::4:  ]",
        "[para(4,5):]",
        "[text(4,5):c:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>a
<blockquote>
<p>b</p>
</blockquote>
<ul>
<li>c</li>
</ul>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_301():
    """
    Test case 301:  This list is tight, because the consecutive block elements are not separated by blank lines:
    """

    # Arrange
    source_markdown = """- a
  > b
  ```
  c
  ```
- d"""
    expected_tokens = [
        "[ulist(1,1):-::2::  \n  \n  ]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > ]",
        "[para(2,5):]",
        "[text(2,5):b:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[fcode-block(3,3):`:3::::::]",
        "[text(4,3):c:]",
        "[end-fcode-block::3:False]",
        "[li(6,1):2::]",
        "[para(6,3):]",
        "[text(6,3):d:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>a
<blockquote>
<p>b</p>
</blockquote>
<pre><code>c
</code></pre>
</li>
<li>d</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_301a():
    """
    Test case 301a:  variation
    """

    # Arrange
    source_markdown = """- a
  >> b
  ```
  c
  ```
- d"""
    expected_tokens = [
        "[ulist(1,1):-::2::  \n  \n  ]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :]",
        "[block-quote(2,4):  :  >> ]",
        "[para(2,6):]",
        "[text(2,6):b:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[end-block-quote:::True]",
        "[fcode-block(3,3):`:3::::::]",
        "[text(4,3):c:]",
        "[end-fcode-block::3:False]",
        "[li(6,1):2::]",
        "[para(6,3):]",
        "[text(6,3):d:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>a
<blockquote>
<blockquote>
<p>b</p>
</blockquote>
</blockquote>
<pre><code>c
</code></pre>
</li>
<li>d</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_301b():
    """
    Test case 301b:  variation
    """

    # Arrange
    source_markdown = """- a
  > b
   ```
   c
   ```
- d"""
    expected_tokens = [
        "[ulist(1,1):-::2::  \n  \n  ]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > ]",
        "[para(2,5):]",
        "[text(2,5):b:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[fcode-block(3,4):`:3::::: :]",
        "[text(4,3):c:\a \a\x03\a]",
        "[end-fcode-block: :3:False]",
        "[li(6,1):2::]",
        "[para(6,3):]",
        "[text(6,3):d:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>a
<blockquote>
<p>b</p>
</blockquote>
<pre><code>c
</code></pre>
</li>
<li>d</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_301cx():
    """
    Test case 301cx:  variation
    """

    # Arrange
    source_markdown = """- a
  > b
      c
- d"""
    expected_tokens = [
        "[ulist(1,1):-::2::  ]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > \n  ]",
        "[para(2,5):\n    ]",
        "[text(2,5):b\nc::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[li(4,1):2::]",
        "[para(4,3):]",
        "[text(4,3):d:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>a
<blockquote>
<p>b
c</p>
</blockquote>
</li>
<li>d</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_301ca():
    """
    Test case 301ca:  variation
    """

    # Arrange
    source_markdown = """- a
  > b
  >    c
- d"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > \n  > ]",
        "[para(2,5):\n   ]",
        "[text(2,5):b\nc::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[li(4,1):2::]",
        "[para(4,3):]",
        "[text(4,3):d:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>a
<blockquote>
<p>b
c</p>
</blockquote>
</li>
<li>d</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_301cb():
    """
    Test case 301cb:  variation
    """

    # Arrange
    source_markdown = """> b
    c
- d"""
    expected_tokens = [
        "[block-quote(1,1)::> \n]",
        "[para(1,3):\n    ]",
        "[text(1,3):b\nc::\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[ulist(3,1):-::2:]",
        "[para(3,3):]",
        "[text(3,3):d:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<blockquote>
<p>b
c</p>
</blockquote>
<ul>
<li>d</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_301d():
    """
    Test case 301d:  variation
    """

    # Arrange
    source_markdown = """- a
  > b
  >
  >     c
- d"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[block-quote(2,3):  :  > \n  >\n  > ]",
        "[para(2,5):]",
        "[text(2,5):b:]",
        "[end-para:::True]",
        "[BLANK(3,4):]",
        "[icode-block(4,9):    :]",
        "[text(4,9):c:]",
        "[end-icode-block:::True]",
        "[end-block-quote:::True]",
        "[li(5,1):2::]",
        "[para(5,3):]",
        "[text(5,3):d:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>a
<blockquote>
<p>b</p>
<pre><code>c
</code></pre>
</blockquote>
</li>
<li>d</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_302():
    """
    Test case 302:  (part 1) A single-paragraph list is tight:
    """

    # Arrange
    source_markdown = """- a"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>a</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_303():
    """
    Test case 303:  (part 2) A single-paragraph list is tight:
    """

    # Arrange
    source_markdown = """- a
  - b"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[ulist(2,3):-::4:  ]",
        "[para(2,5):]",
        "[text(2,5):b:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>a
<ul>
<li>b</li>
</ul>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_304():
    """
    Test case 304:  This list is loose, because of the blank line between the two block elements in the list item:
    """

    # Arrange
    source_markdown = """1. ```
   foo
   ```

   bar"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   \n   \n   ]",
        "[fcode-block(1,4):`:3::::::]",
        "[text(2,4):foo:]",
        "[end-fcode-block::3:False]",
        "[BLANK(4,1):]",
        "[para(5,4):]",
        "[text(5,4):bar:]",
        "[end-para:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>
<pre><code>foo
</code></pre>
<p>bar</p>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_305x():
    """
    Test case 305:  (part 1) Here the outer list is loose, the inner list tight:
    """

    # Arrange
    source_markdown = """* foo
  * bar

  baz"""
    expected_tokens = [
        "[ulist(1,1):*::2::  ]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[ulist(2,3):*::4:  ]",
        "[para(2,5):]",
        "[text(2,5):bar:]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[end-ulist:::True]",
        "[para(4,3):]",
        "[text(4,3):baz:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>
<p>foo</p>
<ul>
<li>bar</li>
</ul>
<p>baz</p>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_305a():
    """
    Test case 305a:  variation on 305
    """

    # Arrange
    source_markdown = """* foo
* foogle    
  * bar

  baz"""
    expected_tokens = [
        "[ulist(1,1):*::2::  ]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[li(2,1):2::]",
        "[para(2,3)::    ]",
        "[text(2,3):foogle:]",
        "[end-para:::True]",
        "[ulist(3,3):*::4:  ]",
        "[para(3,5):]",
        "[text(3,5):bar:]",
        "[end-para:::True]",
        "[BLANK(4,1):]",
        "[end-ulist:::True]",
        "[para(5,3):]",
        "[text(5,3):baz:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>
<p>foo</p>
</li>
<li>
<p>foogle</p>
<ul>
<li>bar</li>
</ul>
<p>baz</p>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_305b():
    """
    Test case 305b:  variation on 305
    """

    # Arrange
    source_markdown = """* foo
 * foogle    
   * bar

   baz"""
    expected_tokens = [
        "[ulist(1,1):*::2::   ]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[li(2,2):3: :]",
        "[para(2,4)::    ]",
        "[text(2,4):foogle:]",
        "[end-para:::True]",
        "[ulist(3,4):*::5:   ]",
        "[para(3,6):]",
        "[text(3,6):bar:]",
        "[end-para:::True]",
        "[BLANK(4,1):]",
        "[end-ulist:::True]",
        "[para(5,4):]",
        "[text(5,4):baz:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>
<p>foo</p>
</li>
<li>
<p>foogle</p>
<ul>
<li>bar</li>
</ul>
<p>baz</p>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_306():
    """
    Test case 306:  (part 2) Here the outer list is loose, the inner list tight:
    """

    # Arrange
    source_markdown = """- a
  - b
  - c

- d
  - e
  - f"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):a:]",
        "[end-para:::True]",
        "[ulist(2,3):-::4:  ]",
        "[para(2,5):]",
        "[text(2,5):b:]",
        "[end-para:::True]",
        "[li(3,3):4:  :]",
        "[para(3,5):]",
        "[text(3,5):c:]",
        "[end-para:::True]",
        "[BLANK(4,1):]",
        "[end-ulist:::True]",
        "[li(5,1):2::]",
        "[para(5,3):]",
        "[text(5,3):d:]",
        "[end-para:::True]",
        "[ulist(6,3):-::4:  ]",
        "[para(6,5):]",
        "[text(6,5):e:]",
        "[end-para:::True]",
        "[li(7,3):4:  :]",
        "[para(7,5):]",
        "[text(7,5):f:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>
<p>a</p>
<ul>
<li>b</li>
<li>c</li>
</ul>
</li>
<li>
<p>d</p>
<ul>
<li>e</li>
<li>f</li>
</ul>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_extra_01x():
    """
    Test case List01:  link definition within a list item
                       copy of test_link_reference_definitions_161 but within list item
    """

    # Arrange
    source_markdown = """- [foo]: /url "title"

  [foo]"""
    expected_tokens = [
        "[ulist(1,1):-::2::  ]",
        '[link-ref-def(1,3):True::foo:: :/url:: :title:"title":]',
        "[BLANK(2,1):]",
        "[para(3,3):]",
        "[link(3,3):shortcut:/url:title::::foo:::::]",
        "[text(3,4):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li><a href="/url" title="title">foo</a></li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_extra_01a():
    """
    Test case Bq03a:  link definition within a list item
                      copy of test_link_reference_definitions_161 but within
                      two distinct list items
    """

    # Arrange
    source_markdown = """- [foo]: /url "title"
- [foo]"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        '[link-ref-def(1,3):True::foo:: :/url:: :title:"title":]',
        "[li(2,1):2::]",
        "[para(2,3):]",
        "[link(2,3):shortcut:/url:title::::foo:::::]",
        "[text(2,4):foo:]",
        "[end-link::]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li></li>
<li><a href="/url" title="title">foo</a></li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_list_items_extra_01b():
    """
    Test case Bq03a:  link definition within a list item
                      copy of test_link_reference_definitions_164 but within
                      a single list item
    """

    # Arrange
    source_markdown = """- [Foo bar]:
  <my url>
  'title'

  [Foo bar]"""
    expected_tokens = [
        "[ulist(1,1):-::2::  \n  \n  ]",
        "[link-ref-def(1,3):True::foo bar:Foo bar:\n:my%20url:<my url>:\n:title:'title':]",
        "[BLANK(4,1):]",
        "[para(5,3):]",
        "[link(5,3):shortcut:my%20url:title::::Foo bar:::::]",
        "[text(5,4):Foo bar:]",
        "[end-link::]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li><a href="my%20url" title="title">Foo bar</a></li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Tests for the extract_whitespace_from_end function.
"""
from pymarkdown.parser_helper import ParserHelper


def test_empty_string():
    """
    Make sure that an empty string is handled properly
    """

    # Arrange
    input_string = ""
    expected_output = (0, "")

    # Act
    actual_output = ParserHelper.extract_whitespace_from_end(input_string)

    # Assert
    assert expected_output == actual_output


def test_string_with_no_trailing_spaces():
    """
    Make sure that a string with no trailing spaces is handled properly
    """

    # Arrange
    input_string = "no trailing spaces"
    expected_output = (len(input_string), "")

    # Act
    actual_output = ParserHelper.extract_whitespace_from_end(input_string)

    # Assert
    assert expected_output == actual_output


def test_string_with_some_trailing_spaces():
    """
    Make sure that a string with some trailing spaces is handled properly
    """

    # Arrange
    input_string = "some trailing spaces  "
    expected_output = (len(input_string) - 2, "  ")

    # Act
    actual_output = ParserHelper.extract_whitespace_from_end(input_string)

    # Assert
    assert expected_output == actual_output



"""
Constants
"""


# pylint: disable=too-few-public-methods
class Constants:
    """
    Constants to use throughout the parser.
    """

    whitespace = "\x20\x09\x0a\x0b\x0c\x0d"

    non_space_whitespace = whitespace[1:]

    ascii_control_characters = (
        "\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f"
        + "\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f"
        + "\x20\x7f"
    )

    unicode_whitespace = "\x20\x09\x0a\x0c\x0d\u00a0\u1680\u2000\u2001\u2002\u2003\u2004\u2005\u2006\u2007\u2008\u2009\u200A\u202F\u205F\u3000"

    punctuation_characters = (
        # standard ascii punctuation
        "\u0020\u0021\u0022\u0023\u0024\u0025\u0026\u0027\u0028"
        + "\u0029\u002a\u002b\u002c\u002d\u002e\u002f"
        + "\u003a\u003b\u003c\u003d\u003e\u003f\u0040"
        + "\u005b\u005c\u005d\u005e\u005f\u0060"
        + "\u007b\u007c\u007d\u007e"
    )
    # http://www.fileformat.info/info/unicode/category/index.htm
    # http://www.fileformat.info/info/unicode/category/P*/list.htm


# pylint: enable=too-few-public-methods



"""
https://github.github.com/gfm/#tables-extension-
"""

from .utils import act_and_assert


def test_tables_extension_198():
    """
    Test case 198:  The delimiter row consists of cells whose only content are hyphens (-), and optionally, a leading or trailing colon (:), or both, to indicate left, right, or center alignment respectively.
    """

    # Arrange
    source_markdown = """| foo | bar |
| --- | --- |
| baz | bim |"""
    expected_tokens = [
        "[para(1,1):\n\n]",
        "[text(1,1):| foo | bar |\n| --- | --- |\n| baz | bim |::\n\n]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>| foo | bar |\n| --- | --- |\n| baz | bim |</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


def test_tables_extension_199():
    """
    Test case 199:  Cells in one column don’t need to match length, though it’s easier to read if they are. Likewise, use of leading and trailing pipes may be inconsistent:
    """

    # Arrange
    source_markdown = """| abc | defghi |
:-: | -----------:
bar | baz"""
    expected_tokens = [
        "[para(1,1):\n\n]",
        "[text(1,1):| abc | defghi |\n:-: | -----------:\nbar | baz::\n\n]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>| abc | defghi |\n:-: | -----------:\nbar | baz</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


def test_tables_extension_200():
    """
    Test case 200:  Include a pipe in a cell’s content by escaping it, including inside other inline spans:
    """

    # Arrange
    source_markdown = """| f\\|oo  |
| ------ |
| b `\\|` az |
| b **\\|** im |"""
    expected_tokens = [
        "[para(1,1):\n\n\n]",
        "[text(1,1):| f\\\b|oo  |\n| ------ |\n| b ::\n\n]",
        "[icode-span(3,5):\\|:`::]",
        "[text(3,9): az |\n| b ::\n]",
        "[emphasis(4,5):2:*]",
        "[text(4,7):\\\b|:]",
        "[end-emphasis(4,9)::]",
        "[text(4,11): im |:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>| f|oo  |\n| ------ |\n| b <code>\\|</code> az |\n| b <strong>|</strong> im |</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


def test_tables_extension_201():
    """
    Test case 201:  (part 1) The table is broken at the first empty line, or beginning of another block-level structure:
    """

    # Arrange
    source_markdown = """| abc | def |
| --- | --- |
| bar | baz |
> bar"""
    expected_tokens = [
        "[para(1,1):\n\n]",
        "[text(1,1):| abc | def |\n| --- | --- |\n| bar | baz |::\n\n]",
        "[end-para:::True]",
        "[block-quote(4,1)::> ]",
        "[para(4,3):]",
        "[text(4,3):bar:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<p>| abc | def |\n| --- | --- |\n| bar | baz |</p>\n<blockquote>\n<p>bar</p>\n</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


def test_tables_extension_202():
    """
    Test case 202:  (part 2) The table is broken at the first empty line, or beginning of another block-level structure:
    """

    # Arrange
    source_markdown = """| abc | def |
| --- | --- |
| bar | baz |
bar

bar"""
    expected_tokens = [
        "[para(1,1):\n\n\n]",
        "[text(1,1):| abc | def |\n| --- | --- |\n| bar | baz |\nbar::\n\n\n]",
        "[end-para:::True]",
        "[BLANK(5,1):]",
        "[para(6,1):]",
        "[text(6,1):bar:]",
        "[end-para:::True]",
    ]
    expected_gfm = (
        """<p>| abc | def |\n| --- | --- |\n| bar | baz |\nbar</p>\n<p>bar</p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


def test_tables_extension_203():
    """
    Test case 203:  The header row must match the delimiter row in the number of cells. If not, a table will not be recognized:
    """

    # Arrange
    source_markdown = """| abc | def |
| --- |
| bar |"""
    expected_tokens = [
        "[para(1,1):\n\n]",
        "[text(1,1):| abc | def |\n| --- |\n| bar |::\n\n]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>| abc | def |\n| --- |\n| bar |</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


def test_tables_extension_204():
    """
    Test case 204:  The remainder of the table’s rows may vary in the number of cells. If there are a number of cells fewer than the number of cells in the header row, empty cells are inserted. If there are greater, the excess is ignored:
    """

    # Arrange
    source_markdown = """| abc | def |
| --- | --- |
| bar |
| bar | baz | boo |"""
    expected_tokens = [
        "[para(1,1):\n\n\n]",
        "[text(1,1):| abc | def |\n| --- | --- |\n| bar |\n| bar | baz | boo |::\n\n\n]",
        "[end-para:::True]",
    ]
    expected_gfm = (
        """<p>| abc | def |\n| --- | --- |\n| bar |\n| bar | baz | boo |</p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


def test_tables_extension_205():
    """
    Test case 205:  If there are no rows in the body, no <tbody> is generated in HTML output:
    """

    # Arrange
    source_markdown = """| abc | def |
| --- | --- |"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):| abc | def |\n| --- | --- |::\n]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>| abc | def |\n| --- | --- |</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
https://github.github.com/gfm/#paragraphs
"""
import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_blank_lines_197():
    """
    Test case 197:  Blank lines at the beginning and end of the document are also ignored.
    """

    # Arrange
    source_markdown = """\a\a

aaa
\a\a

# aaa

  """.replace(
        "\a", " "
    )
    expected_tokens = [
        "[BLANK(1,1):  ]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[text(3,1):aaa:]",
        "[end-para:::True]",
        "[BLANK(4,1):  ]",
        "[BLANK(5,1):]",
        "[atx(6,1):1:0:]",
        "[text(6,3):aaa: ]",
        "[end-atx::]",
        "[BLANK(7,1):]",
        "[BLANK(8,1):  ]",
    ]
    expected_gfm = """<p>aaa</p>
<h1>aaa</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_blank_lines_197a():
    """
    Test case 197a:  Extra blanks to test
    """

    # Arrange
    source_markdown = """\a\a
\a
aaa
""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[BLANK(1,1):  ]",
        "[BLANK(2,1): ]",
        "[para(3,1):]",
        "[text(3,1):aaa:]",
        "[end-para:::True]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<p>aaa</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
https://github.github.com/gfm/#paragraph
"""
import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_paragraph_series_c_b():
    """
    Test case:  Paragraphs ends with a backslash escape
    was:        test_paragraph_extra_21
    """

    # Arrange
    source_markdown = """a fun day\\\\"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a fun day\\\b\\:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a fun day\\</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_c_bh():
    """
    Test case:  Paragraph ends with a backslash as in a hard line break
    was:        test_paragraph_extra_22
    """

    # Arrange
    source_markdown = """this was \\
"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):this was \\:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
    ]
    expected_gfm = """<p>this was \\</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_c_sh():
    """
    Test case:  Paragraph ends with 2+ spaces as in a hard line break
    was:        test_paragraph_extra_23
    """

    # Arrange
    source_markdown = """no line break?\a\a\a
""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[para(1,1)::   ]",
        "[text(1,1):no line break?:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
    ]
    expected_gfm = """<p>no line break?</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_c_cs():
    """
    Test case:  Paragraph string ending with a code span.
    was:        test_paragraph_extra_24
    """

    # Arrange
    source_markdown = """a fun ``day``"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):a fun :]",
        "[icode-span(1,7):day:``::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a fun <code>day</code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_c_cr():
    """
    Test case:  Paragraph string ending with a character reference.
    was:        test_paragraph_extra_25
    """

    # Arrange
    source_markdown = """played on &amp;"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):played on \a&amp;\a\a&\a&amp;\a\a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>played on &amp;</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_c_rh():
    """
    Test case:  Paragraph string ending with a raw html block.
    was:        test_paragraph_extra_26
    """

    # Arrange
    source_markdown = """really, <there it='is'>"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):really, :]",
        "[raw-html(1,9):there it='is']",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>really, <there it='is'></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_c_ua():
    """
    Test case:  Paragraph string ending with an URI autolink
    was:        test_paragraph_extra_27
    """

    # Arrange
    source_markdown = """at <http://www.google.com>"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):at :]",
        "[uri-autolink(1,4):http://www.google.com]",
        "[end-para:::True]",
    ]
    expected_gfm = (
        """<p>at <a href="http://www.google.com">http://www.google.com</a></p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_c_ea():
    """
    Test case:  Paragraph string ending with an email autolink
    was:        test_paragraph_extra_28
    """

    # Arrange
    source_markdown = """contact <foo@bar.com>"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):contact :]",
        "[email-autolink(1,9):foo@bar.com]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>contact <a href="mailto:foo@bar.com">foo@bar.com</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_c_e():
    """
    Test case:  Paragraph string ending with an emphasis
    was:        test_paragraph_extra_29
    """

    # Arrange
    source_markdown = """it's *me*"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):it's :]",
        "[emphasis(1,6):1:*]",
        "[text(1,7):me:]",
        "[end-emphasis(1,9)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>it's <em>me</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_c_l():
    """
    Test case:  Paragraph string ending with a link.
    was:        test_paragraph_extra_30
    """

    # Arrange
    source_markdown = """like [Foo](/uri "t")"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):like :]",
        '[link(1,6):inline:/uri:t::::Foo:False:":: :]',
        "[text(1,7):Foo:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>like <a href="/uri" title="t">Foo</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_c_i():
    """
    Test case:  Paragraph string ending with an image
    was:        test_paragraph_extra_31
    """

    # Arrange
    source_markdown = """an ![foo](/url "t")"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):an :]",
        '[image(1,4):inline:/url:t:foo::::foo:False:":: :]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p>an <img src="/url" alt="foo" title="t" /></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Tests for the find_nth_occurrence function.
"""
from pymarkdown.parser_helper import ParserHelper


def test_find_nth_simple_positive():
    """
    Make sure that...
    """

    # Arrange
    search_in = "abc"
    search_for = "abc"
    nth = 1
    expected_output = 0

    # Act
    actual_output = ParserHelper.find_nth_occurrence(search_in, search_for, nth)

    # Assert
    assert expected_output == actual_output


def test_find_nth_simple_negative():
    """
    Make sure that...
    """

    # Arrange
    search_in = "abc"
    search_for = "bcd"
    nth = 1
    expected_output = -1

    # Act
    actual_output = ParserHelper.find_nth_occurrence(search_in, search_for, nth)

    # Assert
    assert expected_output == actual_output


def test_find_nth_double_positive():
    """
    Make sure that...
    """

    # Arrange
    search_in = "abc def abc"
    search_for = "abc"
    nth = 2
    expected_output = 8

    # Act
    actual_output = ParserHelper.find_nth_occurrence(search_in, search_for, nth)

    # Assert
    assert expected_output == actual_output


def test_find_nth_double_negative():
    """
    Make sure that...
    """

    # Arrange
    search_in = "abc"
    search_for = "abc def def"
    nth = 2
    expected_output = -1

    # Act
    actual_output = ParserHelper.find_nth_occurrence(search_in, search_for, nth)

    # Assert
    assert expected_output == actual_output


def test_find_nth_overlap_positive():
    """
    Make sure that...
    """

    # Arrange
    search_in = "abababababababa"
    search_for = "aba"
    nth = 4
    expected_output = 6

    # Act
    actual_output = ParserHelper.find_nth_occurrence(search_in, search_for, nth)

    # Assert
    assert expected_output == actual_output



"""
https://github.github.com/gfm/#textual-content
"""

import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_textual_content_671():
    """
    Test case 671:  (part 1) Any characters not given an interpretation by the above rules will be parsed as plain textual content.
    """

    # Arrange
    source_markdown = """hello $.;'there"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):hello $.;'there:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>hello $.;'there</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_672():
    """
    Test case 672:  (part 2) Any characters not given an interpretation by the above rules will be parsed as plain textual content.
    """

    # Arrange
    source_markdown = """Foo χρῆν"""
    expected_tokens = ["[para(1,1):]", "[text(1,1):Foo χρῆν:]", "[end-para:::True]"]
    expected_gfm = """<p>Foo χρῆν</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_textual_content_673():
    """
    Test case 673:  Internal spaces are preserved verbatim:
    """

    # Arrange
    source_markdown = """Multiple     spaces"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):Multiple     spaces:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>Multiple     spaces</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
https://github.github.com/gfm/#setext-headings
"""
import pytest

from .utils import act_and_assert


# pylint: disable=too-many-lines
@pytest.mark.gfm
def test_setext_headings_050():
    """
    Test case 050:  Simple examples:
    """

    # Arrange
    source_markdown = """Foo *bar*
=========

Foo *bar*
---------"""
    expected_tokens = [
        "[setext(2,1):=:9::(1,1)]",
        "[text(1,1):Foo :]",
        "[emphasis(1,5):1:*]",
        "[text(1,6):bar:]",
        "[end-emphasis(1,9)::]",
        "[end-setext::]",
        "[BLANK(3,1):]",
        "[setext(5,1):-:9::(4,1)]",
        "[text(4,1):Foo :]",
        "[emphasis(4,5):1:*]",
        "[text(4,6):bar:]",
        "[end-emphasis(4,9)::]",
        "[end-setext::]",
    ]
    expected_gfm = """<h1>Foo <em>bar</em></h1>
<h2>Foo <em>bar</em></h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_051():
    """
    Test case 051:  The content of the heading may span more than one line:
    """

    # Arrange
    source_markdown = """Foo *bar
baz*
===="""
    expected_tokens = [
        "[setext(3,1):=:4::(1,1)]",
        "[text(1,1):Foo :]",
        "[emphasis(1,5):1:*]",
        "[text(1,6):bar\nbaz::\n]",
        "[end-emphasis(2,4)::]",
        "[end-setext::]",
    ]
    expected_gfm = """<h1>Foo <em>bar
baz</em></h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_052():
    """
    Test case 052:  The heading’s raw content is formed by concatenating the lines and removing initial and final whitespace.
    """

    # Arrange
    source_markdown = """  Foo *bar
baz*\t
===="""
    expected_tokens = [
        "[setext(3,1):=:4:  :(1,3):\t]",
        "[text(1,3):Foo :]",
        "[emphasis(1,7):1:*]",
        "[text(1,8):bar\nbaz::\n]",
        "[end-emphasis(2,4)::]",
        "[end-setext::]",
    ]
    expected_gfm = """<h1>Foo <em>bar
baz</em></h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_052a():
    """
    Test case 052a:  Deal with multiple lines that start with whitespace.
    """

    # Arrange
    source_markdown = """  a
  b
  c
==="""
    expected_tokens = [
        "[setext(4,1):=:3:  :(1,3)]",
        "[text(1,3):a\nb\nc::\n  \x02\n  \x02]",
        "[end-setext::]",
    ]
    expected_gfm = """<h1>a
b
c</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_052b():
    """
    Test case 052a:  Deal with multiple lines that start with whitespace.
    """

    # Arrange
    source_markdown = """  a\a
  b\a
  c
===""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[setext(4,1):=:3:  :(1,3)]",
        "[text(1,3):a\nb\nc:: \n  \x02 \n  \x02]",
        "[end-setext::]",
    ]
    expected_gfm = """<h1>a
b
c</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_052c():
    """
    Test case 052a:  Deal with multiple lines that start with whitespace.
    """

    # Arrange
    source_markdown = """  a\a
  b\a\a
  c
===""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[setext(4,1):=:3:  :(1,3)]",
        "[text(1,3):a\nb:: \n  \x02]",
        "[hard-break(2,4):  :\n]",
        "[text(3,3):c::  \x02]",
        "[end-setext::]",
    ]
    expected_gfm = """<h1>a
b<br />
c</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens, show_debug=False)


@pytest.mark.gfm
def test_setext_headings_052d():
    """
    Test case 052a:  Deal with multiple lines that start with whitespace.
    """

    # Arrange
    source_markdown = """  a\a
  *b*\a
  c
===""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[setext(4,1):=:3:  :(1,3)]",
        "[text(1,3):a\n:: \n  \x02]",
        "[emphasis(2,3):1:*]",
        "[text(2,4):b:]",
        "[end-emphasis(2,5)::]",
        "[text(2,6):\nc:: \n  \x02]",
        "[end-setext::]",
    ]
    expected_gfm = """<h1>a
<em>b</em>
c</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_052ex():
    """
    Test case 052a:  Deal with multiple lines that start with whitespace.
    """

    # Arrange
    source_markdown = """  a\a
  *b*\a\a
  c
===""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[setext(4,1):=:3:  :(1,3)]",
        "[text(1,3):a\n:: \n  \x02]",
        "[emphasis(2,3):1:*]",
        "[text(2,4):b:]",
        "[end-emphasis(2,5)::]",
        "[hard-break(2,6):  :\n]",
        "[text(3,3):c::  \x02]",
        "[end-setext::]",
    ]
    expected_gfm = """<h1>a
<em>b</em><br />
c</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_052ea():
    """
    Test case 052a:  Deal with multiple lines that start with whitespace.
    """

    # Arrange
    source_markdown = """  a\a
  *b*\a\a
  a\a
  *b*\a\a
  c
===""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[setext(6,1):=:3:  :(1,3)]",
        "[text(1,3):a\n:: \n  \x02]",
        "[emphasis(2,3):1:*]",
        "[text(2,4):b:]",
        "[end-emphasis(2,5)::]",
        "[hard-break(2,6):  :\n]",
        "[text(3,3):a\n::  \x02 \n  \x02]",
        "[emphasis(4,3):1:*]",
        "[text(4,4):b:]",
        "[end-emphasis(4,5)::]",
        "[hard-break(4,6):  :\n]",
        "[text(5,3):c::  \x02]",
        "[end-setext::]",
    ]
    expected_gfm = """<h1>a
<em>b</em><br />
a
<em>b</em><br />
c</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_052eb():
    """
    Test case 052a:  Deal with multiple lines that start with whitespace.
    """

    # Arrange
    source_markdown = """ a\a
 *b*\a\a
  a\a
  *b*\a\a
   a\a
   *b*\a\a
c
===""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[setext(8,1):=:3: :(1,2)]",
        "[text(1,2):a\n:: \n \x02]",
        "[emphasis(2,2):1:*]",
        "[text(2,3):b:]",
        "[end-emphasis(2,4)::]",
        "[hard-break(2,5):  :\n]",
        "[text(3,3):a\n::  \x02 \n  \x02]",
        "[emphasis(4,3):1:*]",
        "[text(4,4):b:]",
        "[end-emphasis(4,5)::]",
        "[hard-break(4,6):  :\n]",
        "[text(5,4):a\n::   \x02 \n   \x02]",
        "[emphasis(6,4):1:*]",
        "[text(6,5):b:]",
        "[end-emphasis(6,6)::]",
        "[hard-break(6,7):  :\n]",
        "[text(7,1):c:]",
        "[end-setext::]",
    ]
    expected_gfm = """<h1>a
<em>b</em><br />
a
<em>b</em><br />
a
<em>b</em><br />
c</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_052f():
    """
    Test case 052f:  Deal with multiple lines that start with whitespace.
    """

    # Arrange
    source_markdown = """  a\a
   *b*\a\a
 c
===""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[setext(4,1):=:3:  :(1,3)]",
        "[text(1,3):a\n:: \n   \x02]",
        "[emphasis(2,4):1:*]",
        "[text(2,5):b:]",
        "[end-emphasis(2,6)::]",
        "[hard-break(2,7):  :\n]",
        "[text(3,2):c:: \x02]",
        "[end-setext::]",
    ]
    expected_gfm = """<h1>a
<em>b</em><br />
c</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_053():
    """
    Test case 053:  The underlining can be any length:
    """

    # Arrange
    source_markdown = """Foo
-------------------------

Foo
="""
    expected_tokens = [
        "[setext(2,1):-:25::(1,1)]",
        "[text(1,1):Foo:]",
        "[end-setext::]",
        "[BLANK(3,1):]",
        "[setext(5,1):=:1::(4,1)]",
        "[text(4,1):Foo:]",
        "[end-setext::]",
    ]
    expected_gfm = """<h2>Foo</h2>
<h1>Foo</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_054():
    """
    Test case 054:  The heading content can be indented up to three spaces, and need not line up with the underlining:
    """

    # Arrange
    source_markdown = """   Foo
---

  Foo
-----

  Foo
  ==="""
    expected_tokens = [
        "[setext(2,1):-:3:   :(1,4)]",
        "[text(1,4):Foo:]",
        "[end-setext::]",
        "[BLANK(3,1):]",
        "[setext(5,1):-:5:  :(4,3)]",
        "[text(4,3):Foo:]",
        "[end-setext::]",
        "[BLANK(6,1):]",
        "[setext(8,3):=:3:  :(7,3)]",
        "[text(7,3):Foo:]",
        "[end-setext:  :]",
    ]
    expected_gfm = """<h2>Foo</h2>
<h2>Foo</h2>
<h1>Foo</h1>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_055():
    """
    Test case 055:  Four spaces indent is too much:
    """

    # Arrange
    source_markdown = """    Foo
    ---

    Foo
---"""
    expected_tokens = [
        "[icode-block(1,5):    :\n    \n\n    ]",
        "[text(1,5):Foo\n---\n\x03\nFoo:]",
        "[end-icode-block:::False]",
        "[tbreak(5,1):-::---]",
    ]
    expected_gfm = """<pre><code>Foo
---

Foo
</code></pre>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_056():
    """
    Test case 056:  The setext heading underline can be indented up to three spaces, and may have trailing spaces:
    """

    # Arrange
    source_markdown = """Foo
   ----      """
    expected_tokens = [
        "[setext(2,4):-:4::(1,1)]",
        "[text(1,1):Foo:]",
        "[end-setext:   :      ]",
    ]
    expected_gfm = """<h2>Foo</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_057():
    """
    Test case 057:  Four spaces is too much:
    """

    # Arrange
    source_markdown = """Foo
    ---"""
    expected_tokens = [
        "[para(1,1):\n    ]",
        "[text(1,1):Foo\n---::\n]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>Foo
---</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_058():
    """
    Test case 058:  The setext heading underline cannot contain internal spaces:
    """

    # Arrange
    source_markdown = """Foo
= =

Foo
--- -"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):Foo\n= =::\n]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[para(4,1):]",
        "[text(4,1):Foo:]",
        "[end-para:::False]",
        "[tbreak(5,1):-::--- -]",
    ]
    expected_gfm = """<p>Foo
= =</p>
<p>Foo</p>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_059():
    """
    Test case 059:  Trailing spaces in the content line do not cause a line break:
    """

    # Arrange
    source_markdown = """Foo\a\a
-----""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[setext(2,1):-:5::(1,1):  ]",
        "[text(1,1):Foo:]",
        "[end-setext::]",
    ]
    expected_gfm = """<h2>Foo</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_060():
    """
    Test case 060:  Nor does a backslash at the end:
    """

    # Arrange
    source_markdown = """Foo\\
----"""
    expected_tokens = [
        "[setext(2,1):-:4::(1,1)]",
        "[text(1,1):Foo\\:]",
        "[end-setext::]",
    ]
    expected_gfm = """<h2>Foo\\</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_061():
    """
    Test case 061:  Since indicators of block structure take precedence over indicators of inline structure, the following are setext headings:
    """

    # Arrange
    source_markdown = """`Foo
----
`

<a title="a lot
---
of dashes"/>"""
    expected_tokens = [
        "[setext(2,1):-:4::(1,1)]",
        "[text(1,1):`Foo:]",
        "[end-setext::]",
        "[para(3,1):]",
        "[text(3,1):`:]",
        "[end-para:::True]",
        "[BLANK(4,1):]",
        "[setext(6,1):-:3::(5,1)]",
        '[text(5,1):\a<\a&lt;\aa title=\a"\a&quot;\aa lot:]',
        "[end-setext::]",
        "[para(7,1):]",
        '[text(7,1):of dashes\a"\a&quot;\a/\a>\a&gt;\a:]',
        "[end-para:::True]",
    ]
    expected_gfm = """<h2>`Foo</h2>
<p>`</p>
<h2>&lt;a title=&quot;a lot</h2>
<p>of dashes&quot;/&gt;</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_062():
    """
    Test case 062:  (part a) The setext heading underline cannot be a lazy continuation line in a list item or block quote:
    """

    # Arrange
    source_markdown = """> Foo
---"""
    expected_tokens = [
        "[block-quote(1,1)::> ]",
        "[para(1,3):]",
        "[text(1,3):Foo:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[tbreak(2,1):-::---]",
    ]
    expected_gfm = """<blockquote>
<p>Foo</p>
</blockquote>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_063():
    """
    Test case 063:  (part b) The setext heading underline cannot be a lazy continuation line in a list item or block quote:
    """

    # Arrange
    source_markdown = """> foo
bar
==="""
    expected_tokens = [
        "[block-quote(1,1)::> \n\n]",
        "[para(1,3):\n\n]",
        "[text(1,3):foo\nbar\n===::\n\n]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p>foo
bar
===</p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_064():
    """
    Test case 064:  (part c) The setext heading underline cannot be a lazy continuation line in a list item or block quote:
    """

    # Arrange
    source_markdown = """- Foo
---"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):Foo:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[tbreak(2,1):-::---]",
    ]
    expected_gfm = """<ul>
<li>Foo</li>
</ul>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_064a():
    """
    Test case 064a:  064 with with other underline
    """

    # Arrange
    source_markdown = """- Foo
---"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):Foo:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[tbreak(2,1):-::---]",
    ]
    expected_gfm = """<ul>
<li>Foo</li>
</ul>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_065():
    """
    Test case 065:  A blank line is needed between a paragraph and a following setext heading, since otherwise the paragraph becomes part of the heading’s content:
    """

    # Arrange
    source_markdown = """Foo
Bar
---"""
    expected_tokens = [
        "[setext(3,1):-:3::(1,1)]",
        "[text(1,1):Foo\nBar::\n]",
        "[end-setext::]",
    ]
    expected_gfm = """<h2>Foo
Bar</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_066():
    """
    Test case 066:  A blank line is needed between a paragraph and a following setext heading, since otherwise the paragraph becomes part of the heading’s content:
    """

    # Arrange
    source_markdown = """---
Foo
---
Bar
---
Baz"""
    expected_tokens = [
        "[tbreak(1,1):-::---]",
        "[setext(3,1):-:3::(2,1)]",
        "[text(2,1):Foo:]",
        "[end-setext::]",
        "[setext(5,1):-:3::(4,1)]",
        "[text(4,1):Bar:]",
        "[end-setext::]",
        "[para(6,1):]",
        "[text(6,1):Baz:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<hr />
<h2>Foo</h2>
<h2>Bar</h2>
<p>Baz</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_067():
    """
    Test case 067:  Setext headings cannot be empty:
    """

    # Arrange
    source_markdown = """
===="""
    expected_tokens = [
        "[BLANK(1,1):]",
        "[para(2,1):]",
        "[text(2,1):====:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>====</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_068():
    """
    Test case 068:  (part a) Setext heading text lines must not be interpretable as block constructs other than paragraphs. So, the line of dashes in these examples gets interpreted as a thematic break:
    """

    # Arrange
    source_markdown = """---
---"""
    expected_tokens = ["[tbreak(1,1):-::---]", "[tbreak(2,1):-::---]"]
    expected_gfm = """<hr />
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_069():
    """
    Test case 069:  (part b) Setext heading text lines must not be interpretable as block constructs other than paragraphs. So, the line of dashes in these examples gets interpreted as a thematic break:
    """

    # Arrange
    source_markdown = """- foo
-----"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[tbreak(2,1):-::-----]",
    ]
    expected_gfm = """<ul>
<li>foo</li>
</ul>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_069a():
    """
    Test case 069a:  069 with other underline
    """

    # Arrange
    source_markdown = """- foo
-----"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[tbreak(2,1):-::-----]",
    ]
    expected_gfm = """<ul>
<li>foo</li>
</ul>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_070():
    """
    Test case 070:  (part c) Setext heading text lines must not be interpretable as block constructs other than paragraphs. So, the line of dashes in these examples gets interpreted as a thematic break:
    """

    # Arrange
    source_markdown = """    foo
---"""
    expected_tokens = [
        "[icode-block(1,5):    :]",
        "[text(1,5):foo:]",
        "[end-icode-block:::False]",
        "[tbreak(2,1):-::---]",
    ]
    expected_gfm = """<pre><code>foo
</code></pre>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_071():
    """
    Test case 071:  (part d) Setext heading text lines must not be interpretable as block constructs other than paragraphs. So, the line of dashes in these examples gets interpreted as a thematic break:
    """

    # Arrange
    source_markdown = """> foo
-----"""
    expected_tokens = [
        "[block-quote(1,1)::> ]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
        "[tbreak(2,1):-::-----]",
    ]
    expected_gfm = """<blockquote>
<p>foo</p>
</blockquote>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_072():
    """
    Test case 072:  If you want a heading with > foo as its literal text, you can use backslash escapes:
    """

    # Arrange
    source_markdown = """\\> foo
------"""
    expected_tokens = [
        "[setext(2,1):-:6::(1,1)]",
        "[text(1,1):\\\b\a>\a&gt;\a foo:]",
        "[end-setext::]",
    ]
    expected_gfm = """<h2>&gt; foo</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_073():
    """
    Test case 073:  Authors who want interpretation 1 can put a blank line after the first paragraph:
    https://github.github.com/gfm/#example-73
    """

    # Arrange
    source_markdown = """Foo

bar
---
baz"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):Foo:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[setext(4,1):-:3::(3,1)]",
        "[text(3,1):bar:]",
        "[end-setext::]",
        "[para(5,1):]",
        "[text(5,1):baz:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>Foo</p>
<h2>bar</h2>
<p>baz</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_074():
    """
    Test case 074:  Authors who want interpretation 2 can put blank lines around the thematic break,
    https://github.github.com/gfm/#example-74
    """

    # Arrange
    source_markdown = """Foo
bar

---

baz"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):Foo\nbar::\n]",
        "[end-para:::True]",
        "[BLANK(3,1):]",
        "[tbreak(4,1):-::---]",
        "[BLANK(5,1):]",
        "[para(6,1):]",
        "[text(6,1):baz:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>Foo
bar</p>
<hr />
<p>baz</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_075():
    """
    Test case 075:  or use a thematic break that cannot count as a setext heading underline, such as
    https://github.github.com/gfm/#example-75
    """

    # Arrange
    source_markdown = """Foo
bar
* * *
baz"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):Foo\nbar::\n]",
        "[end-para:::False]",
        "[tbreak(3,1):*::* * *]",
        "[para(4,1):]",
        "[text(4,1):baz:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>Foo
bar</p>
<hr />
<p>baz</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_setext_headings_076():
    """
    Test case 076:  Authors who want interpretation 3 can use backslash escapes:
    https://github.github.com/gfm/#example-76
    """

    # Arrange
    source_markdown = """Foo
bar
\\---
baz"""
    expected_tokens = [
        "[para(1,1):\n\n\n]",
        "[text(1,1):Foo\nbar\n\\\b---\nbaz::\n\n\n]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>Foo
bar
---
baz</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)
